;; Generated package description from ellama.el  -*- no-byte-compile: t -*-
(define-package "ellama" "0.8.13" "Tool for interacting with LLMs" '((emacs "28.1") (llm "0.6.0") (spinner "1.7.4")) :commit "b94d5952d4e36e1caeef5ac0b3ecd6d89009082b" :authors '(("Sergey Kostyaev" . "sskostyaev@gmail.com")) :maintainer '("Sergey Kostyaev" . "sskostyaev@gmail.com") :keywords '("help" "local" "tools") :url "http://github.com/s-kostyaev/ellama")
